//---------------------------------------------------------------------------
#ifdef WIN32
#include <windows.h>
#endif
//---------------------------------------------------------------------------
#define CLAMP(m,M,v) (((v) < (m)) ? (m) : ( ((v) > (M)) ? (M) : (v)  ))
#include <math.h>
//---------------------------------------------------------------------------
#include <GL/glu.h>
#include <GL/gl.h>

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include "CException.h"
#include "CTexture.h"
//---------------------------------------------------------------------------
#ifndef NO_JPG

#include "CTextureJPG.h"

#endif
//---------------------------------------------------------------------------

#include "CTextureTGA.h"

//---------------------------------------------------------------------------

#ifndef NO_PPM
#include "CTexturePPM.h"

#endif
//---------------------------------------------------------------------------

#include "OSAdapter.h"
//---------------------------------------------------------------------------
CTexture::CTexture()
{
  m_iWidth=0;
  m_iHeight=0;
  m_Data=NULL;
  m_bLoaded=false;
  m_bAlpha=false;
  m_szName[0]='\0';
  //  m_bLuminance=false;
  m_bDataOwner=true;
}
//---------------------------------------------------------------------------
CTexture::CTexture(const char *name,int w,int h,bool alpha,unsigned char *data)
{
  m_iWidth=w;
  m_iHeight=h;
  m_Data=data;
  m_bLoaded=true;
  m_bAlpha=alpha;
  strncpy(m_szName,name,TEXTURE_NAME_SIZE);
  m_bDataOwner=false;
}
//---------------------------------------------------------------------------
CTexture::CTexture(const CTexture& tex)
{
  m_iWidth=tex.getWidth();
  m_iHeight=tex.getHeight();
  strcpy(m_szName,tex.getName());
  if (tex.isLoaded())
  {
    m_Data=new unsigned char[tex.memsize()];
    memcpy(m_Data,tex.getData(),tex.memsize());
    m_bLoaded=true;
    m_bAlpha=isAlpha();
  }
  else
  {
    m_Data=NULL;
    m_bLoaded=false;
  }
  m_bDataOwner=true;
}
//---------------------------------------------------------------------------
CTexture::CTexture(CTexture *ptex)
{
  m_iWidth=ptex->getWidth();
  m_iHeight=ptex->getHeight();
  strcpy(m_szName,ptex->getName());
  if (ptex->isLoaded())
  {
    m_Data=new unsigned char[ptex->memsize()];
    memcpy(m_Data,ptex->getData(),ptex->memsize());
    m_bLoaded=true;
    m_bAlpha=isAlpha();
  }
  else
  {
    m_Data=NULL;
    m_bLoaded=false;
  }
  m_bDataOwner=true;
}
//---------------------------------------------------------------------------
CTexture::~CTexture()
{
  if (m_bDataOwner)
    unload();
}
//---------------------------------------------------------------------------
void CTexture::unload()
{
  if (!m_bDataOwner)
    throw CException("Cannot unload a reference texture: it is not the texture object that has loaded data !");
  if (m_bLoaded)
  {
    delete[] (m_Data);
    m_bLoaded=false;
  }
}
//---------------------------------------------------------------------------
void  CTexture::convert2Alpha(unsigned char r,
			      unsigned char g,
			      unsigned char b,
			      unsigned char tolerence)
{
  computeAlpha(CTexture::alpha_color(r,g,b,tolerence));
}
//---------------------------------------------------------------------------
void CTexture::loadAlphaMap(const char *n)
{
  CTexture *tex=loadTexture(n);
  loadAlphaMap(tex);
  delete (tex);
}
//---------------------------------------------------------------------------
void CTexture::loadAlphaMap(CTexture *altex)
{
  unsigned char *ad=altex->getData();
  int t;

  if (altex->getWidth() != getWidth() || altex->getHeight() != getHeight())
    throw CException("Alpha map and texture should have same size !");
  if (altex->isAlpha())
    t=4;
  else
    t=3;
  if (!isAlpha())
    convert2Alpha(0,0,0,0);
  for (int i=0;i<getWidth();i++)
  {
    for (int j=0;j<getHeight();j++)
    {
      // use red as alpha
      m_Data[(i+j*getWidth())*4+3]=ad[(i+j*getWidth())*t];
    }
  }
  m_bAlpha=true;
}
//---------------------------------------------------------------------------
void CTexture::computeAlpha(const CTexture::alpha_functor& af)
{
  int i,t=3;
  unsigned char *data;
	
  if (!isLoaded())
    load();
  if (isAlpha())
    t=4;
  data=new unsigned char[m_iHeight*m_iWidth*4];
  for (i=0;i<m_iHeight*m_iWidth;i++)
  {
    data[i*4]=m_Data[i*t];
    data[i*4+1]=m_Data[i*t+1];
    data[i*4+2]=m_Data[i*t+2];
	data[i*4+3]=af.computeAlpha(m_Data[i*t],m_Data[i*t+1],m_Data[i*t+2]);
  }
  delete [](m_Data);
  m_bAlpha=true;
  m_Data=data;
}
//---------------------------------------------------------------------------
void CTexture::convertToNormal(double scale)
{
  unsigned char *d=computeNormals(scale);
  delete (m_Data);
  m_Data=d;
}
// --------------------------------------------------------
unsigned char *CTexture::computeNormals(double scale)
{
  int            t,i,j,sz;
  vertex         du,dv,n;
  double         c0,cu,cv;
  unsigned char *data;
  int w,h;

  w=m_iWidth;
  h=m_iHeight;
  t=isAlpha()?4:3;
  data=new unsigned char[t*w*h];
  
  sz=w*h*t;
  for (i=0;i<w;i++)
  {
    for (j=0;j<h;j++)
    {
      c0=( ((double)m_Data[((i+j*w)*t) % sz])
	   +((double)m_Data[((i+j*w)*t+1) % sz])
	   +((double)m_Data[((i+j*w)*t+2) % sz]))/3.0;
      cu=( ((double)m_Data[(((i+1)+j*w)*t) % sz])
	   +((double)m_Data[(((i+1)+j*w)*t+1) % sz])
	   +((double)m_Data[(((i+1)+j*w)*t+2) % sz]))/3.0;
      cv=( ((double)m_Data[((i+(j+1)*w)*t) % sz])
	   +((double)m_Data[((i+(j+1)*w)*t+1) % sz])
	   +((double)m_Data[((i+(j+1)*w)*t+2) % sz]))/3.0;
      du=vertex(1.0,
		0.0,
		(c0-cu)*scale/255.0);
      dv=vertex(0.0,
		1.0,
		(c0-cv)*scale/255.0);
      n=du.cross(dv);
      n.normalize();
      data[((i+j*w)*t  )%sz]=(unsigned char)CLAMP(0,255,
					     (128.0 + n.x*127.0));
      data[((i+j*w)*t+1)%sz]=(unsigned char)CLAMP(0,255,
					     (128.0 + n.y*127.0));
      data[((i+j*w)*t+2)%sz]=(unsigned char)CLAMP(0,255,
					     (128.0 + n.z*127.0));
      if (isAlpha())
	data[((i+j*w)*t+3)%sz]=(unsigned char)m_Data[((i+j*w)*t+3) % sz];
    }
  }
  return (data);
}
//---------------------------------------------------------------------------
int CTexture::memsize() const
{
  return ((isAlpha()?4:3)*m_iWidth*m_iHeight);
}
//---------------------------------------------------------------------------
void CTexture::flipY()
{
  int           i,j;
  unsigned char pix[4];
  int           t=isAlpha()?4:3;

  for (i=0;i<m_iWidth;i++)
  {
    for (j=0;j<m_iHeight/2;j++)
    {
      memcpy(pix,&(m_Data[(i+j*m_iWidth)*t]),t);
      memcpy(&(m_Data[(i+j*m_iWidth)*t]),&(m_Data[(i+(m_iHeight-j-1)*m_iWidth)*t]),t);
      memcpy(&(m_Data[(i+(m_iHeight-j-1)*m_iWidth)*t]),pix,t);
    }
  }
}
//---------------------------------------------------------------------------
void CTexture::flipX()
{
  int           i,j;
  unsigned char pix[4];
  int           t=isAlpha()?4:3;

  for (i=0;i<m_iWidth/2;i++)
  {
    for (j=0;j<m_iHeight;j++)
    {
      memcpy(pix,&(m_Data[(i+j*m_iWidth)*t]),t);
      memcpy(&(m_Data[(i+j*m_iWidth)*t]),&(m_Data[((m_iWidth-i-1)+j*m_iWidth)*t]),t);
      memcpy(&(m_Data[((m_iWidth-i-1)+j*m_iWidth)*t]),pix,t);
    }
  }
}
//---------------------------------------------------------------------------
CTexture *CTexture::loadTexture(const char *n)
{
  char *nos=OSAdapter::convertName(n);
  int l=strlen(nos);
  const char *ext=nos+l-4;
  if (strstr(nos,".tga") == ext)
  {
    return (new CTextureTGA(nos));
  }
#ifndef NO_JPG

  else if (strstr(nos,".jpg") == ext)

  {

    return (new CTextureJPG(nos));

  }

#endif

#ifndef NO_PPM

  else if (strstr(nos,".ppm") == ext)
  {
    return (new CTexturePPM(nos));
  }

#endif
  else
    throw CException(true,"libTexture - (load) Unknown file type (%s)",nos);
  return (NULL);
}
//---------------------------------------------------------------------------
void CTexture::saveTexture(CTexture *tex,const char *n)
{
  char *nos=OSAdapter::convertName(n);
  int l=strlen(nos);
  const char *ext=nos+l-4;
  if (strstr(nos,".tga") == ext)
  {
    CTextureTGA *tmp=new CTextureTGA(n,tex->getWidth(),tex->getHeight(),tex->isAlpha(),tex->getData());
    tmp->save();
    delete (tmp);
  } 
  else
    throw CException(true,"libTexture - (save) Unknown file type (%s)",nos);    
}
//---------------------------------------------------------------------------
