// ----------------------------------------------------
#ifdef WIN32
#  include <windows.h>
#  define M_PI 3.141592653589793
#endif
#include <GL/glu.h>
#include "CMobile.h"
#include "CWorld.h"
#include "messages_CMobile.h"
// ----------------------------------------------------
#include "CInterface.h"
// ----------------------------------------------------
#include <string>
#include <sstream>
#include "antlr/AST.hpp"
#include "CalcWalker.hpp"
#include "CalcParser.hpp"
#include "CalcLexer.hpp"
#include "MyASTNode.h"  // $FAB
int cur_line=0; // pour recuperer ligne d'erreur depuis calc.g
int cur_mob=0;  // ... et la transmettre a CException

// ----------------------------------------------------
#include "CIconFactory.h"
// ----------------------------------------------------
#include <qbrush.h>
#include <qcolor.h>
#include <qtextedit.h>
//#include <qstatusbar.h>
// ----------------------------------------------------
PERSIST_ID(CMobile);
// -----------------
PERSIST_BEGIN_MEMBERS(CMobile);
PERSIST_MEMBER(m_Pos);
PERSIST_MEMBER(m_Speed);
PERSIST_MEMBER(m_iIcon);
PERSIST_MEMBER(m_dAngle);
PERSIST_MEMBER(m_uiState);
PERSIST_MEMBER(m_uiColor);
PERSIST_MEMBER(m_dWidth);
PERSIST_MEMBER(m_dHeight);
PERSIST_MEMBER(m_bExport);
PERSIST_MEMBER(m_bVisible);
PERSIST_MEMBER(m_sExprs);
PERSIST_MEMBER(m_sExprsCollide);
PERSIST_MEMBER(m_sExprsStart);
PERSIST_TABLE(m_sExprsBorder,0,3);
PERSIST_TABLE(m_Vars,0,MAX_VAR-1);
PERSIST_MEMBER(m_iRank);
PERSIST_END_MEMBERS;
// ----------------------------------------------------
ANTLR_USING_NAMESPACE(std);
ANTLR_USING_NAMESPACE(antlr);
// ----------------------------------------------------
#define SHOW_CAPSULE 0
// ----------------------------------------------------
CMobile::CMobile()
{
  init();
  m_uiState=MOBILE_STATE_STOPPED;
}
//-------------------------------------------------------------------
void CMobile::init()
{
  m_dAngle=0;
  m_uiColor=0xFFFFFFFF;
  m_dLength=1.0;
  m_dRadius=1.0;
  m_dWidth=1.0;
  m_dHeight=1.0;
  m_uiState=0;
  m_iIcon=-1;
  m_bContact=false;
  m_bContactFirst=false;
  m_iExprLastError=0;
  m_iExprLineError=0;
  m_iRank=-1;

  for (int i=0; i<=MAX_VAR; i++)
    m_Vars[i]=0.0;
  
  m_iUpdateRate=10;
  m_iLastTime=(rand() % m_iUpdateRate)+SDL_GetTicks();

  m_iGhostMasterType=0;
  m_iUpdateMode=NTWRK_FAST;

  m_ParserExprs=NULL;
  m_ParserExprsStart=NULL;
  m_ParserExprsCollide=NULL;
  m_ParserExprsBorder[0]=NULL;
  m_ParserExprsBorder[1]=NULL;
  m_ParserExprsBorder[2]=NULL;
  m_ParserExprsBorder[3]=NULL;
  m_bValid=false;
  m_bExport=false;
  m_bVisible=true;
  m_bEditing=false;
  m_bStarting = false;
}
//-------------------------------------------------------------------
CMobile::CMobile(int t,int id,
		 const unsigned char *d,int l)
  : IGhost(t,id,d,l)
{
  m_iIcon=-1;
  m_bExport=false;
  m_bVisible=true;
  m_bContact=false;
  m_bContactFirst=false;
  m_dLength=1.0;
  m_dRadius=1.0;
  m_dWidth=1.0;
  m_dHeight=1.0;
  m_bStarting = false;
  m_ParserExprs=NULL;
  m_ParserExprsStart=NULL;
  m_ParserExprsCollide=NULL;
  m_ParserExprsBorder[0]=NULL;
  m_ParserExprsBorder[1]=NULL;
  m_ParserExprsBorder[2]=NULL;
  m_ParserExprsBorder[3]=NULL;
  readNetworkData(d,l);
  PERSIST.unregister(this);
}
// ----------------------------------------------------
CMobile::CMobile(CMobile *m)
{
  m_iIcon=m->getIcon();
  m_Pos=m->getPos();
  m_Speed=m->getSpeed();
  m_dAngle=m->m_dAngle;
  m_uiColor=m->getColor();
  m_dLength=m->getLength();
  m_dRadius=m->getRadius();
  m_uiState=m->m_uiState;
  m_dWidth=m->m_dWidth;
  m_dHeight=m->m_dHeight;
  m_bContact=m->m_bContact;
  m_bContactFirst=m->m_bContactFirst;
  for (int i=0; i<=MAX_VAR; i++)
    m_Vars[i]=m->m_Vars[i];

  m_iUpdateRate=m->m_iUpdateRate;
  m_iLastTime=(rand() % m_iUpdateRate)+SDL_GetTicks();

  m_iGhostMasterType=0;
  m_iUpdateMode=NTWRK_FAST;

  m_ParserExprs=NULL;
  m_ParserExprsStart=NULL;
  m_ParserExprsCollide=NULL;
  m_ParserExprsBorder[0]=NULL;
  m_ParserExprsBorder[1]=NULL;
  m_ParserExprsBorder[2]=NULL;
  m_ParserExprsBorder[3]=NULL;

  m_sExprs=m->getExprs();
  m_sExprsCollide=m->getExprsCollide();
  m_sExprsBorder[0]=m->getExprsBorder(0);
  m_sExprsBorder[1]=m->getExprsBorder(1);
  m_sExprsBorder[2]=m->getExprsBorder(2);
  m_sExprsBorder[3]=m->getExprsBorder(3);
  m_sExprsStart=m->getExprsStart();

  m_bExport=m->isExported();
  m_bVisible=m->isVisible();

  m_iRank=-1;
  m_bStarting = false;
  
  compile();
}
// ----------------------------------------------------
CMobile::~CMobile()
{
  if (m_ParserExprs != NULL)
    delete (m_ParserExprs);
  if (m_ParserExprsStart != NULL)
    delete (m_ParserExprsStart);
  if (m_ParserExprsCollide != NULL)
    delete (m_ParserExprsCollide);
  for (int b=0;b<4;b++)
    if (m_ParserExprsBorder[b] != NULL)
      delete (m_ParserExprsBorder[b]);
}
// ----------------------------------------------------
void CMobile::reset()
{
//  m_dAngle=0.0;
//  m_uiColor=0xF0F0F0FF;
//  m_dLength=1.0;
//  m_dRadius=1.0;
//  m_dWidth=1.0;
//  m_dHeight=1.0;
  m_uiState=0;
  m_Pos=CVertex(0,0,0);
  m_Speed=CVertex(0,0,0);
  m_bValid=false;
  for (int i=0; i<=MAX_VAR; i++)
    m_Vars[i]=0.0;
  m_uiState=MOBILE_STATE_STOPPED;
  m_bExport=false;
  m_bVisible=true;
  updateExport();
  updateVisible();
/* 
   m_iIcon=-1;
   m_sExprs.resize(0);
   m_sExprsCollide.resize(0);
   m_sExprsBorder[0].resize(0);
   m_sExprsBorder[1].resize(0);
   m_sExprsBorder[2].resize(0);
   m_sExprsBorder[3].resize(0);
   m_sExprsStart.resize(0);

*/
  compile();
}
// ----------------------------------------------------
void CMobile::networkUpdate()
{
  int tm=SDL_GetTicks();

  if (isMaster() && isExported() && isVisible())
  {
    if (tm - m_iLastTime > m_iUpdateRate)
    {
      setNeedUpdate(true);
      m_iLastTime=tm;
    }
  }
}
// ----------------------------------------------------
void CMobile::draw()
{
  static GLUquadric *q=gluNewQuadric();
  CVertex p=getPos();

  if (!isVisible())
    return;
  // adjust collision volume to icon
  if (m_iIcon >= 0)
    {
      m_dRadius=m_dWidth*CIconFactory::getIcon(m_iIcon)->get_radius();
      m_dLength=m_dHeight*CIconFactory::getIcon(m_iIcon)->get_length();
    }
  else
    {
      m_dRadius=m_dWidth;
      m_dLength=m_dHeight;
    }

  // draw
  glPushAttrib(GL_ENABLE_BIT);

  glPushMatrix();
  glTranslated(p.x(),p.y(),p.z());
  
	//bug if m_dAngle >= 10.0E17
  glRotated(m_dAngle*180.0/M_PI,0.0,0.0,1.0);

  if (m_iIcon >= 0)
    {
      glColor3d(((m_uiColor & 0xFF000000) >> 24)/255.0,
		((m_uiColor & 0x00FF0000) >> 16)/255.0,
		((m_uiColor & 0x0000FF00) >>  8)/255.0);
      try
	{
	  CIconFactory::getIcon(m_iIcon)->draw(m_dHeight,m_dWidth); // $FAB
	}
      catch (...)
	{
      
	}
#if SHOW_CAPSULE
      glDisable(GL_DEPTH_TEST);
      glEnable(GL_BLEND);
      glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
      glColorMask(GL_TRUE,GL_FALSE,GL_FALSE,GL_FALSE); // $FAB
      glColor4d(1.0,0.0,0.0,0.7);
      glRotated(CIconFactory::getIcon(m_iIcon)->get_angle0()*180.0/M_PI,0.0,0.0,1.0); // $FAB
#endif
    }
  else
    {
      glDisable(GL_BLEND);
      glColor3d(((m_uiColor & 0xFF000000) >> 24)/255.0,
		((m_uiColor & 0x00FF0000) >> 16)/255.0,
		((m_uiColor & 0x0000FF00) >>  8)/255.0);
      glPushMatrix();
      glTranslated(0.0,-m_dLength/2.0,0.0);
      gluSphere(q,m_dRadius,10,10);
      glTranslated(0.0,m_dLength,0.0);
      gluSphere(q,m_dRadius,10,10);
      glPopMatrix();
  
      glPushMatrix();  
      glTranslated(0.0,m_dLength/2.0,0.0);
      glRotated(90.0,1.0,0.0,0.0);
      gluCylinder(q,m_dRadius,m_dRadius,m_dLength,10,10);
      glPopMatrix();
#if SHOW_CAPSULE
      glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE); // $FAB
#endif
    }
  glPopMatrix();
  glPopAttrib();
}
//---------------------------------------------------------------
void CMobile::compile()
{
  static ASTFactory factory("myastnode",MyASTNode::factory);

  display_error_clear(); cur_line=0;
  m_bValid=true;
  m_iExprLastError = -1;
  m_iExprLineError = 0;
  try
  {
    // motion
    {
      istringstream is(m_sExprs.str()+"\n");
      CalcLexer lexer(is);
      lexer.setFilename("[ERROR] ");
      if( m_ParserExprs != NULL )
        delete m_ParserExprs;
      m_ParserExprs = new CalcParser(lexer);
      m_ParserExprs->initializeASTFactory(factory);
      m_ParserExprs->setASTFactory(&factory);
      //      m_ParserExprs->setASTNodeFactory(&(MyASTNode::factory)); // $FAB
      m_ParserExprs->setFilename("[ERROR] ");
      m_iExprLastError++;
      m_ParserExprs->start();
      /*
      if( m_ParserExprs->getAST() )
      cout << endl << m_ParserExprs->getAST()->toStringList() << endl;
      */
    }
    // start
    {
      istringstream is(m_sExprsStart.str()+"\n");
      CalcLexer lexer(is);
      lexer.setFilename("[ERROR] ");
      if( m_ParserExprsStart != NULL )
        delete m_ParserExprsStart;
      m_ParserExprsStart = new CalcParser(lexer);
      m_ParserExprsStart->initializeASTFactory(factory);
      m_ParserExprsStart->setASTFactory(&factory);
      //      m_ParserExprsStart->setASTNodeFactory(&MyASTNode::factory); // $FAB
      m_ParserExprsStart->setFilename("[ERROR] ");
      m_iExprLastError++;
      m_ParserExprsStart->start();
    }
    // collide
    {
      istringstream is(m_sExprsCollide.str()+"\n");
      CalcLexer lexer(is);
      lexer.setFilename("[ERROR] ");
      if( m_ParserExprsCollide != NULL )
        delete m_ParserExprsCollide;
      m_ParserExprsCollide = new CalcParser(lexer);
      m_ParserExprsCollide->initializeASTFactory(factory);
      m_ParserExprsCollide->setASTFactory(&factory);
      //      m_ParserExprsCollide->setASTNodeFactory(&MyASTNode::factory); // $FAB
      m_ParserExprsCollide->setFilename("[ERROR] ");
      m_iExprLastError++;
      m_ParserExprsCollide->start();
    }
    // borders 
    for (int b=0;b<4;b++)
    {
      istringstream is(m_sExprsBorder[b].str()+"\n");
      CalcLexer lexer(is);
      lexer.setFilename("[ERROR] ");
      if( m_ParserExprsBorder[b] != NULL )
        delete (m_ParserExprsBorder[b]);
      m_ParserExprsBorder[b] = new CalcParser(lexer);
      m_ParserExprsBorder[b]->initializeASTFactory(factory);
      m_ParserExprsBorder[b]->setASTFactory(&factory);
      //      m_ParserExprsBorder[b]->setASTNodeFactory(&MyASTNode::factory); // $FAB
      m_ParserExprsBorder[b]->setFilename("[ERROR] ");
      m_iExprLastError++;
      m_ParserExprsBorder[b]->start();
    }
  }
  catch (RecognitionException& e)
    {
     m_bValid=false;
     m_iExprLineError=e.getLine();
     display_error(e.getMessage(),cur_mob,e.getLine(),e.getColumn());
    }
  catch (CException& e)
  {
    m_bValid=false;
    m_iExprLineError=e.getLine();
    display_error(e.getMsg(),cur_mob,e.getLine());
  }
  catch (...)
  {
    m_bValid=false;
    m_iExprLineError=cur_line+1;
    display_error("Erreur de syntaxe !");
  }
  g_MainWindow->updateDisplay(false);
}
//---------------------------------------------------------------
void CMobile::update()
{
  if (NULL == m_ParserExprs || !isValid() || isStopped())
    return;
  try
  {
    CalcWalker walker;
    walker.bindTo(this, FIELD_MVT);
    walker.start(m_ParserExprs->getAST());
  }
  catch (RecognitionException& e)
    {
     m_bValid=false;
     m_iExprLineError=e.getLine();
     display_error(e.getMessage(),cur_mob,e.getLine(),e.getColumn());
    }
  catch (CException& e)
  {
    m_bValid=false;
    m_iExprLastError=0;
    m_iExprLineError=e.getLine();
    display_error(e.getMsg(),cur_mob,e.getLine());
  }
  catch (...)
  {
    m_bValid=false;
    m_iExprLineError=cur_line+1;
    display_error("Erreur de syntaxe !");
  }
}
//---------------------------------------------------------------
void CMobile::startParserForField(void* w, int i) 
{
  try
  {
    CalcWalker* walker = (CalcWalker*)(w);
    m_iExprLineError=0;
    switch(i) 
    {
    case FIELD_MVT:     walker->start(m_ParserExprs->getAST());          m_iExprLastError=0; break;
    case FIELD_START:   walker->start(m_ParserExprsStart->getAST());     m_iExprLastError=1; break;
    case FIELD_COLLIDE: walker->start(m_ParserExprsCollide->getAST());   m_iExprLastError=2; break;
    case FIELD_BORDER0: walker->start(m_ParserExprsBorder[0]->getAST()); m_iExprLastError=3+WORLD_BORDER_BOTTOM; break;
    case FIELD_BORDER1: walker->start(m_ParserExprsBorder[1]->getAST()); m_iExprLastError=3+WORLD_BORDER_RIGHT; break;
    case FIELD_BORDER2: walker->start(m_ParserExprsBorder[2]->getAST()); m_iExprLastError=3+WORLD_BORDER_LEFT; break;
    case FIELD_BORDER3: walker->start(m_ParserExprsBorder[3]->getAST()); m_iExprLastError=3+WORLD_BORDER_TOP; break;
    }
  }
  catch (CException& e)
  {
    m_bValid=false;
    m_iExprLineError=e.getLine();
    display_error(e.getMsg(),cur_mob,e.getLine());
  }
}
//---------------------------------------------------------------
void CMobile::collide(CMobile *c,const CVertex& dir)
{
  if (NULL == m_ParserExprsCollide || !isValid() || isStopped())
    return;
  if (!m_bContact)
  {
    m_bContactFirst=true;
  }
  else
  {
    m_bContactFirst=false;
  }
  try
  {
    CalcWalker walker;
    walker.bindTo(this, FIELD_COLLIDE, c,dir);
    walker.start(m_ParserExprsCollide->getAST());
  }
  catch (CException& e)
  {
    m_bValid=false;
    m_iExprLastError=2;
    m_iExprLineError=e.getLine();
    display_error(e.getMsg(),cur_mob,e.getLine());
  }
}
//---------------------------------------------------------------
void CMobile::start()
{
  if (NULL == m_ParserExprsStart || !isValid())
    return;
  if (m_bStarting) 
  {
    display_warning(MSG_restart,cur_mob,cur_line+1);
    return;
  }
  m_bStarting = true; // pour tester si recursif
  try
  {
    CalcWalker walker;
    walker.bindTo(this, FIELD_START);
    walker.start(m_ParserExprsStart->getAST());
  }
  catch (CException& e)
  {
    m_bValid=false;
    m_iExprLastError=1;
    m_iExprLineError=e.getLine();
    display_error(e.getMsg(),cur_mob,e.getLine());
  }
  m_bStarting = false;
}
//---------------------------------------------------------------
void CMobile::setEditing(bool b)
{
  m_bEditing=b;
  /*
    if (m_bEditing && !isStopped())
    {
    setStopped(true);
    g_MainWindow->updateDisplay(false);
    }
  */
}
//---------------------------------------------------------------
void CMobile::collideBorder(int b)
{
  if (NULL == m_ParserExprsBorder[b] || !isValid() || isStopped())
    return;
  try
  {
    CalcWalker walker;
    walker.bindTo(this, FIELD_BORDER0+b);
    walker.start(m_ParserExprsBorder[b]->getAST());
  }
  catch (CException& e)
  {
    m_bValid=false;
    m_iExprLastError=3+b;
    m_iExprLineError=e.getLine();
    display_error(e.getMsg(),cur_mob,e.getLine());
  }
}
//---------------------------------------------------------------
void CMobile::updateExport()
{
  if (!isMaster())
    return;
  if (m_bExport)
    sendMessageToGhosts(MOBILE_MSG_EXPORT_TRUE,NULL,0);
  else
    sendMessageToGhosts(MOBILE_MSG_EXPORT_FALSE,NULL,0);
}
//---------------------------------------------------------------
void CMobile::updateVisible()
{
  if (!isMaster())
    return;
  if (m_bVisible)
    sendMessageToGhosts(MOBILE_MSG_VISIBLE_TRUE,NULL,0);
  else
    sendMessageToGhosts(MOBILE_MSG_VISIBLE_FALSE,NULL,0);
}
//---------------------------------------------------------------
void CMobile::execStop()
{
  if (!isMaster())
    sendMessageToMaster(MOBILE_MSG_EXEC_STOP,NULL,0);
  else
    setStopped(true);
  if (g_MainWindow->selected() == this)
    g_MainWindow->updateDisplay(false);
}
//---------------------------------------------------------------
void CMobile::execStart()
{
  if (!isMaster())
    sendMessageToMaster(MOBILE_MSG_EXEC_START,NULL,0);
  else
    setStopped(false);
  if (g_MainWindow->selected() == this)
    g_MainWindow->updateDisplay(false);
}
//---------------------------------------------------------------
void CMobile::execRestart()
{
  if (!isMaster()) 
  {
    sendMessageToMaster(MOBILE_MSG_EXEC_RESTART,NULL,0);
  }
  else 
  {  
    setStopped(true);
    setStopped(false);
  }
  if (g_MainWindow->selected() == this)
    g_MainWindow->updateDisplay(false);
}
//---------------------------------------------------------------
void CMobile::execMoveTo(double x,double y)
{
  mobile_moveto_msg m;

  if (!isMaster())
  {
    m.x=x;
    m.y=y;
    sendMessageToMaster(MOBILE_MSG_EXEC_MOVETO,(unsigned char *)&m,sizeof(mobile_moveto_msg));
  }
  else
    m_Pos=CVertex(x,y);
  if (g_MainWindow->selected() == this)
    g_MainWindow->updateDisplay(false);
}
//---------------------------------------------------------------
void CMobile::setStopped(bool b) 
{
  bool last=isStopped();

  if (b) 
    m_uiState=(m_uiState |  MOBILE_STATE_STOPPED); 
  else   
    m_uiState=(m_uiState & ~MOBILE_STATE_STOPPED);
  if (!b && last)
    start();
}
//---------------------------------------------------------------
//***************************************************************************
// Network
//***************************************************************************
//---------------------------------------------------------------
// methods from INetworkObject
//---------------------------------------------------------------
int CMobile::networkDataSize()
{
  return (sizeof(mobile_data));
}
//---------------------------------------------------------------
const unsigned char *CMobile::writeNetworkData()
{
  static mobile_data m;
  
  m.pos[0]=m_Pos.x();
  m.pos[1]=m_Pos.y();
  m.speed[0]=m_Speed.x();
  m.speed[1]=m_Speed.y();
  m.angle=(unsigned char)((int)(m_dAngle*255.0/(2.0*M_PI)) & 255);
  // bug with angle value -> will be change on v1.2 (not yet cause insert network incompatibility)
  //m.angle= m_dAngle/(2.0*M_PI);
  m.color=m_uiColor;
  m.w=m_dWidth;
  m.h=m_dHeight;
  m.icon=m_iIcon;
  return (reinterpret_cast<unsigned char *>(&m));
}
//---------------------------------------------------------------
void CMobile::readNetworkData(const unsigned char *d,int l)
{
  const mobile_data *m;
  
  m=reinterpret_cast<const mobile_data *>(d);
  m_Pos=CVertex(m->pos[0],m->pos[1],0.0);
  m_Speed=CVertex(m->speed[0],m->speed[1],0.0);
  m_dAngle=((double)m->angle)*(2.0*M_PI)/255.0;
  // bug with angle value -> will be change on v1.2 (not yet cause insert network incompatibility)
  //m_dAngle= m->angle*(2.0*M_PI);
  m_uiColor=m->color;
  m_dWidth=m->w;
  m_dHeight=m->h;
  m_iIcon=m->icon;
  // on ne transmet pas les m_Vars !
  for (int i=0; i<=MAX_VAR; i++)
    m_Vars[i]=0.0;
}
//---------------------------------------------------------------
void CMobile::onGhostMasterDied()
{
  cerr << "Ghost master died " << getGhostId() << endl;
  WORLD.removeMobile(this);
}
//---------------------------------------------------------------
void CMobile::messageFromMaster(ng_ghost_msg *m,const unsigned char *d,int l)
{
  switch (m->m_iInfo)
  {
  case MOBILE_MSG_EXPORT_FALSE:
    m_bExport=false;
    break;
  case MOBILE_MSG_EXPORT_TRUE:
    m_bExport=true;
    break;
  case MOBILE_MSG_VISIBLE_FALSE:
    m_bVisible=false;
    break;
  case MOBILE_MSG_VISIBLE_TRUE:
    m_bVisible=true;
    break;
  }
}
//---------------------------------------------------------------
void CMobile::messageFromGhost(int id,const ng_ghost_msg *m,
			       const unsigned char *d,int l)
{
  //  cerr << "MESSAGE FROM GHOST " << m->m_iInfo << endl;
  switch (m->m_iInfo)
  {
  case MOBILE_MSG_EXEC_STOP:
    setStopped(true);
    break;
  case MOBILE_MSG_EXEC_START:
    setStopped(false);
    break;
  case MOBILE_MSG_EXEC_RESTART:
    setStopped(true);
    setStopped(false);
    break;
  case MOBILE_MSG_EXEC_MOVETO:
    {
      const mobile_moveto_msg *m=reinterpret_cast<const mobile_moveto_msg *>(d);
      m_Pos=CVertex(m->x,m->y);
      break;
    }
  }
}
//---------------------------------------------------------------
