#include <GL/glut.h>
#include <math.h>
#include<stdio.h> // pour debug
#include "icones.h"
#include "CTexture.h"

//   CC icones.C -c -I ../include/

// === icones.C ===========================================
// --- classes generales ------

GLuint stdmesh::initlist() { if (!list) list=glGenLists(1); return list;  }
void   stdmesh::freelist() { if (list) { glDeleteLists(list, 1); list=0;} }
GLuint stdmesh::inittext() { if (!text) glGenTextures(1,&text); return text; }
void   stdmesh::freetext() { if (text) { glDeleteTextures(1,&text); text=0; }}

void stdmesh::draw(double sh, double sr) {
  glPushMatrix();
  glScalef(sr,sh,sr); // sr,sh,1.
  if (list) glCallList(list);
  glPopMatrix();
}

void stdmesh::begininit(const char* tex, const char* alpha) {
  if (tex) 
  {                // on a besoin d'une texture
    if (inittext()) 
    {           // init id
      CTexture *img = CTexture::loadTexture(tex); // load from disk
      if (img) 
      {
	if (alpha) 
	{   // il y a une texture de mask (i.e. alpha separe')
	  CTexture *img2 = CTexture::loadTexture(alpha);
	  if (img2) 
	  {
	    img->loadAlphaMap(img2); // on l'ajoute a l'image
	    delete img2;
	  }
	}
	type = (img->isAlpha());
	glBindTexture(GL_TEXTURE_2D,text);
	if (img->isAlpha())
	  gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA,     // load to board
			    img->getWidth(), img->getHeight(), 
			    GL_RGBA, GL_UNSIGNED_BYTE, img->getData());
	else
	  gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB,
			    img->getWidth(), img->getHeight(), 
			    GL_RGB, GL_UNSIGNED_BYTE, img->getData());
	// img->createGlId(text);
	delete img;
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
			GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glBindTexture(GL_TEXTURE_2D,0);
      }
      else // image pas lisible
	text=0;
    }
  }

  if (initlist()) {         // init id
    glNewList(list, GL_COMPILE);                 // cree display-list
    glPushAttrib(GL_ENABLE_BIT|GL_TEXTURE_BIT);  // sauve gc
    if (text) {
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D,text);         // ev, active texture
      if (type) {
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	//glBlendEquation(GL_FUNC_ADD);
        glEnable(GL_ALPHA_TEST);
   	glAlphaFunc(GL_GEQUAL,0.02);
      }
    }
    // glEnable(GL_RESCALE_NORMAL);
    glPushMatrix();
    // glScalef(100./WORLD_SCALE,100./WORLD_SCALE,100./WORLD_SCALE);
    glScalef(ICON_SCALE,ICON_SCALE,ICON_SCALE);
  }
}  
  
void stdmesh::endinit() {
  if (list) {
    glPopMatrix();
    glPopAttrib();                               // restore gc
    glEndList();
  }
}

image::image(const char* name, const char* tex,const char* alpha):stdmesh(name) {
  begininit(tex,alpha);
  glTranslatef(0,0,1);
  glNormal3f(0,0,1);
  glBegin(GL_QUADS);
  glTexCoord2f(0,1);
  glVertex2f(-1,-1);
  glTexCoord2f(1,1);
  glVertex2f( 1,-1);
  glTexCoord2f(1,0);
  glVertex2f( 1, 1);
  glTexCoord2f(0,0);
  glVertex2f(-1, 1);
  glEnd();
  endinit();
}
/*
void image::draw(double sh=1., double sr=1.) {
  glPushMatrix();
  glScalef(sr*get_radius(),sh*get_height(),1.);
  if (list) glCallList(list);
  glPopMatrix();
}
*/

imagepure::imagepure(const char* name, const char* tex,const char* alpha):stdmesh(name) {
  begininit(tex,alpha);
  glColor3f(1,1,1);      // <- seule diff: on garde couleur d'origine
  glTranslatef(0,0,1);
  glNormal3f(0,0,1);
  glBegin(GL_QUADS);
  glTexCoord2f(0,1);
  glVertex2f(-1,-1);
  glTexCoord2f(1,1);
  glVertex2f( 1,-1);
  glTexCoord2f(1,0);
  glVertex2f( 1, 1);
  glTexCoord2f(0,0);
  glVertex2f(-1, 1);
  glEnd();
  endinit();
}

planet::planet(const char* name, const char* tex,const char* alpha):stdmesh(name) {
  begininit(tex,alpha);
  glColor3f(1,1,1);
  glTranslatef(0,0,1);
  glEnable(GL_TEXTURE_GEN_S);
  glEnable(GL_TEXTURE_GEN_T);
  glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
  glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
  GLfloat paramU[4] = {.5,0,0,.5};
  GLfloat paramV[4] = {0,.5,0,.5};
  glTexGenfv(GL_S, GL_OBJECT_PLANE, paramU);
  glTexGenfv(GL_T, GL_OBJECT_PLANE, paramV);
  glutSolidSphere(1.,20,20);
  endinit();
}

// ---  utilitaires ------

void mySolidCylinder(GLdouble radius, GLdouble height,
		     GLint slices, GLint stacks) {
  int i,j;
  float z0,z1,a;
  
  for (j=0; j<stacks; j++) {
    z0 = (2*j-stacks)*height/stacks; z1 = z0+2*height/stacks;
    glBegin(GL_QUAD_STRIP);
    for (i=0,a=0; i<slices; i++,a+=6.283185/slices) {
      glNormal3f(cos(a),sin(a),0);
      glVertex3f(radius*cos(a),radius*sin(a),z0);
      glNormal3f(cos(a),sin(a),0);
      glVertex3f(radius*cos(a),radius*sin(a),z1);
    }
    a=0;
    glVertex3f(radius*cos(a),radius*sin(a),z0);
    glVertex3f(radius*cos(a),radius*sin(a),z1);
    glEnd();
  }
}

// ---  classes particulieres ------

sphere::sphere():stdmesh("sphere") {
  begininit();
  glTranslatef(0,0,1);
  glutSolidSphere(1.,20,20);
  endinit();
}

balle::balle():stdmesh("balle") {
  begininit();
  glTranslatef(0,0,1);
  glRotatef(90,1,0,0);
  glutSolidSphere(1.,10,10);
  glColor3f(0,0,0);
  glPolygonOffset(-1,1);
  glEnable(GL_POLYGON_OFFSET_LINE);
  glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
  glutSolidSphere(1.,10,10);
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  glDisable(GL_POLYGON_OFFSET_LINE);
  endinit();
}

rond::rond():stdmesh("rond") {
  int slices=30;
  double radius=1.;
  int i; double a;
  begininit();
  glTranslatef(0,0,0.05);
  glNormal3f(0,0,1);
  glBegin(GL_TRIANGLE_FAN);
  glVertex3f(0,0,0);
  for (i=0,a=0; i<slices; i++,a+=6.283185/slices)
     glVertex3f(radius*cos(a),radius*sin(a),0);
  a=0;
  glVertex3f(radius*cos(a),radius*sin(a),0);
  glEnd();

  endinit();
}

spot::spot():stdmesh("spot") {
  int slices=30;
  double radius=1.;
  int i; double a;
  begininit();
  glTranslatef(0,0,0.05);
  glDisable(GL_LIGHTING);
  glBegin(GL_TRIANGLE_FAN);
  glVertex3f(0,0,0);
  for (i=0,a=0; i<slices; i++,a+=6.283185/slices)
     glVertex3f(radius*cos(a),radius*sin(a),0);
  a=0;
  glVertex3f(radius*cos(a),radius*sin(a),0);
  glEnd();

  endinit();
}

cercle::cercle():stdmesh("cercle") {
  int slices=30;
  double radius=1.;
  int i; double a;
  begininit();
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  //glBlendEquation(GL_FUNC_ADD);
  glEnable(GL_LINE_SMOOTH);
  glLineWidth(4.);
  glTranslatef(0,0,0.05);
  glNormal3f(0,0,1);
  glBegin(GL_LINE_LOOP);
  for (i=0,a=0; i<slices; i++,a+=6.283185/slices)
     glVertex3f(radius*cos(a),radius*sin(a),0);
  glEnd();
  glLineWidth(1.);

  endinit();
}

volant::volant():stdmesh("volant") {
  int slices=30;
  double radius=1.;
  int i; double a;
  begininit();
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  //glBlendEquation(GL_FUNC_ADD);
  glDepthMask(GL_FALSE);
  glEnable(GL_LINE_SMOOTH);
  glLineWidth(4.);
  glTranslatef(0,0,0.05);
  glNormal3f(0,0,1);
  glBegin(GL_LINE_LOOP);
  for (i=0,a=0; i<slices; i++,a+=6.283185/slices)
     glVertex3f(radius*cos(a),radius*sin(a),0);
  glEnd();
  glLineWidth(8.);
  glBegin(GL_LINES);
  glVertex3f(0,0,0);
  glVertex3f(0,1,0);
  glVertex3f(0,0,0);
  glVertex3f(-.866,-.5,0);
  glVertex3f(0,0,0);
  glVertex3f(.866,-.5,0);
  glEnd();
  glLineWidth(1.);
  glDepthMask(GL_TRUE);
  endinit();
}

cube::cube():stdmesh("cube",1.2*ICON_SCALE,1.2*ICON_SCALE) {
  begininit();
  glTranslatef(0,0,1);
  glutSolidCube(2.);
  endinit();
}

brique::brique():stdmesh("brique",1.1*ICON_SCALE,.55*ICON_SCALE,1.5708) {
  begininit();
  glTranslatef(0,0,.5);
  glScalef(2.,1.,.5);
  glutSolidCube(1.);
  glColor3f(0,0,0);
  glPolygonOffset(-1,1);
  glEnable(GL_POLYGON_OFFSET_LINE);
  glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
  glutSolidCube(1.);
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  glDisable(GL_POLYGON_OFFSET_LINE);
  endinit();
}

but::but(double l, double h, double p,char * dataDir):stdmesh("but",l*ICON_SCALE,p*ICON_SCALE) { 
  double pp=p*.5,hh=h*.8;
  begininit((std::string(dataDir)+std::string("/../data/but.ppm")).c_str(),(std::string(dataDir)+std::string("/../data/but.mask.ppm")).c_str());
  // begininit("../data/but_noa.ppm");
  glBegin(GL_QUADS);
  glNormal3f(0,0,1); // dessus
  glTexCoord2f(.5,1);
  glVertex3f(-pp,-l,hh);
  glTexCoord2f(1,1);
  glVertex3f( p,-l,h);
  glTexCoord2f(1,0);
  glVertex3f( p, l,h);
  glTexCoord2f(.5,0);
  glVertex3f(-pp, l,hh);
  glNormal3f(0,-1,0); // cote' droit
  glTexCoord2f(0,.5);
  glVertex3f(-p,-l,0);
  glTexCoord2f(1,.5);
  glVertex3f( p,-l,0);
  glTexCoord2f(1,1);
  glVertex3f( p,-l,h);
  glTexCoord2f(0.,1);
  // glTexCoord2f(.5*(1-pp/p),1-.5*(1-hh/h)); // .25 .9
  glVertex3f(-pp, -l,hh);
  glNormal3f(0,1,0); // cote' gauche
  glTexCoord2f(0,.5);
  glVertex3f(-p, l,0);
  glTexCoord2f(1,.5);
  glVertex3f( p, l,0);
  glTexCoord2f(1,1);
  glVertex3f( p, l,h);
  glTexCoord2f(0,1);
  glVertex3f(-pp, l,hh);
  glNormal3f(-1,0,0); // arriere
  glTexCoord2f(0,0.5);
  glVertex3f(-p, l,0);
  glTexCoord2f(1,0.5);
  glVertex3f(-p,-l,0);
  glTexCoord2f(1,1);
  glVertex3f(-pp,-l,hh);
  glTexCoord2f(0,1);
  glVertex3f(-pp, l,hh);
  glEnd();
  endinit();
}

raquette::raquette(char * dataDir):stdmesh("raquette",1.5*ICON_SCALE,.7*ICON_SCALE) {
  begininit((std::string(dataDir)+std::string("/../data/raquette.ppm")).c_str(),(std::string(dataDir)+std::string("/../data/raquette.mask.ppm")).c_str());
  // begininit("../data/raquette.mask.ppm");
  glTranslatef(0,-.6,1);
  glScalef(2.,2.,2.);
  glRotatef(45.,0,1,0);
  glNormal3f(0,0,1);
  glBegin(GL_QUADS);
  glTexCoord2f(0,1);
  glVertex2f(-1,-1);
  glTexCoord2f(1,1);
  glVertex2f( 1,-1);
  glTexCoord2f(1,0);
  glVertex2f( 1, 1);
  glTexCoord2f(0,0);
  glVertex2f(-1, 1);
  glEnd();
  endinit();
}

teapot::teapot():stdmesh("theiere",ICON_SCALE,.7*ICON_SCALE,1.5708) { 
  begininit();
  glTranslatef(0,.1,1);
  // glRotatef(90,1,0,0);
  glutSolidTeapot(1.);
  endinit();
}

void triangle::draw(double sh, double sr) {
  sr *= 1.3*get_radius(); sh *= 1.3*get_height(); // <><>
  glNormal3f(0,0,1.);
  glBegin(GL_TRIANGLES);
  glVertex3f(-sr,-.5*sh,.2);
  glVertex3f( sr,-.5*sh,.2);
  glVertex3f( 0.,(1.732-.5)*sh,.2);
  glEnd();
}

void laser::draw(double sh, double sr) {
 glPushAttrib(GL_ENABLE_BIT|GL_TEXTURE_BIT);  // sauve gc
 glPushMatrix();
 sr *= get_radius();
 sh *= get_height();
 glScalef(sr,sh,1.);
 // glEnable(GL_BLEND);
 // glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
 // //glBlendEquation(GL_FUNC_ADD);
 // glEnable(GL_LINE_SMOOTH);
 // glLineWidth(4.);
 glDisable(GL_LIGHTING);
 glColor3f(1,.3,0);
 glTranslatef(0,0,1);
 glBegin(GL_QUADS);
 glVertex2f(-1,-1); // .1,1
 glVertex2f( 1,-1);
 glVertex2f( 1, 1);
 glVertex2f(-1, 1);
 glEnd();
 // glLineWidth(1.);
 glPopMatrix();
 glPopAttrib();
}

void trait::draw(double sh, double sr) {
 sh *= get_height();
 glPushAttrib(GL_ENABLE_BIT|GL_TEXTURE_BIT);  // sauve gc
 glEnable(GL_BLEND);
 glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
 //glBlendEquation(GL_FUNC_ADD);
 glEnable(GL_LINE_SMOOTH);
 glLineWidth(2.*sr);
 glDisable(GL_LIGHTING);
 glTranslatef(0,0,1);
 glBegin(GL_LINES);
 glVertex3f(0,-sh,1);
 glVertex3f(0, sh,1);
 glEnd();
 glLineWidth(1.);
 glPopAttrib();
}

void ressort::draw(double sh, double sr) {
 int i; float dx;
 sh *= get_height();
 glPushAttrib(GL_ENABLE_BIT|GL_TEXTURE_BIT);  // sauve gc
 glEnable(GL_BLEND);
 glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
 //glBlendEquation(GL_FUNC_ADD);
 glEnable(GL_LINE_SMOOTH);
 glLineWidth(2);
 glDisable(GL_LIGHTING);
 glTranslatef(0,0,1);
 glBegin(GL_LINE_STRIP);
 glVertex3f(0,-sh,1);
 glVertex3f(0,-sh*.5*(1+1/6.),1);
 for (i=-3,dx=-sr*5; i<=3; i++,dx=-dx)
   glVertex3f(dx, sh*.5*i/3., 1);
 glVertex3f(0, sh*.5*(1+1/6.),1);
 glVertex3f(0, sh,1);
 glEnd();
 glLineWidth(1.);
 glPopAttrib();
}

void aiguille::draw(double sh, double sr) {
 sh *= get_height();
 glPushAttrib(GL_ENABLE_BIT|GL_TEXTURE_BIT);  // sauve gc
 glPushMatrix();
 glEnable(GL_BLEND);
 glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
 //glBlendEquation(GL_FUNC_ADD);
 glEnable(GL_LINE_SMOOTH);
 glLineWidth(2.*sr);
 glDisable(GL_LIGHTING);
 glScalef(sh,sh,1);
 glTranslatef(0,0,1);
 glBegin(GL_LINES);
 glVertex2f(0, 0);
 glVertex2f(0, 1);
 glVertex2f(-.25,.7);
 glVertex2f(0, 1);
 glVertex2f(.25,.7);
 glVertex2f(0, 1);
 glEnd();
 glLineWidth(1.);
 glPopMatrix();
 glPopAttrib();
}

paroie::paroie(double l):stdmesh("paroie",l*ICON_SCALE,ICON_SCALE) {
  l -= get_radius()/ICON_SCALE;
  begininit();
  glTranslatef(0,0,1);
  glRotatef(90,1,0,0);
  mySolidCylinder(1.,l,20,1);
  glTranslatef(0,0,-l);
  glutSolidSphere(1.,10,10);
  glTranslatef(0,0,2*l);
  glutSolidSphere(1.,10,10);
  endinit();
}

void paroie2::draw(double sh, double sr) {
  float r = get_radius()*sr;
  float l = get_height()*sh-r;
  glPushMatrix();
  glTranslatef(0,0,1);
  glRotatef(90,1,0,0);
  mySolidCylinder(r,l,20,1);
  glTranslatef(0,0,-l);
  glutSolidSphere(r,10,10);
  glTranslatef(0,0,2*l);
  glutSolidSphere(r,10,10);
  glPopMatrix();
}

soucoupe::soucoupe():stdmesh("soucoupe",1.4*ICON_SCALE,1.1*ICON_SCALE) {
  begininit();
  glTranslatef(0,0,1);
  // corps
  glPushMatrix();
  glScalef(1.,1.5,.2);
  glutSolidSphere(1.,20,20);
  glPopMatrix();
  // reacteurs
  glPushMatrix();
  glTranslatef(-.8,-1,.4);
  glScalef(.2,1.,.2);
  glutSolidSphere(1.,20,20);
  glPopMatrix();
  glPushMatrix();
  glTranslatef(+.8,-1,.4);
  glScalef(.2,1.,.2);
  glutSolidSphere(1.,20,20);
  glPopMatrix();
  glBegin(GL_TRIANGLES);
  // queue verticale
  glNormal3f(.7,0,.7);
  glVertex3f(0,-1.,.3);
  glVertex3f(0.1,-1.,.2);
  glVertex3f(0,.5,.2);
  glNormal3f(-.7,0,.7);
  glVertex3f(0,-1.,.3);
  glVertex3f(-0.1,-1.,.2);
  glVertex3f(0,.5,.2);
  // queue horizontale
  glColor3f(.3,.3,.3);
  //glNormal3f(0,0,1);
  //glVertex3f(-.8,-1.,.4);
  //glVertex3f( .8,-1.,.4);
  //glVertex3f(0,.5,.2);
  glNormal3f(.2,0,1);
  glVertex3f(-.8,-1.2,.5);
  glVertex3f(  0,-1.0,.2);
  glVertex3f(0,.5,.2);
  glNormal3f(-.2,0,1);
  glVertex3f(.8,-1.2,.5);
  glVertex3f( 0,-1.0,.2);
  glVertex3f(0,.5,.2);
  glEnd();
  endinit();
}

fusee::fusee():stdmesh("fusee",1.3*ICON_SCALE,.3*ICON_SCALE) {
  begininit();
  glTranslatef(0,-.3,1);
  // corps
  glPushMatrix();
  glRotatef(-90,1,0,0);
  glutSolidCone(.3,2,20,1);
  glPushMatrix();
  glRotatef(180,0,1,0);
  glutSolidCone(.3,.6,20,1);
  glPopMatrix();
  glPopMatrix();
  // 3 reacteurs
  double l = .6;
  double a = 35.; // 22
  glTranslatef(0,.1,0);
  glPushMatrix();
  glRotatef(90.,0,1,0);
  glRotatef(a,0,0,1);
  glTranslatef(0,-l,0);
  glScalef(.2,1.,.2);
  glutSolidSphere(l,20,20);
  glPopMatrix();
  glPushMatrix(); // #2
  glRotatef(90.+120.,0,1,0);
  glRotatef(a,0,0,1);
  glTranslatef(0,-l,0);
  glScalef(.2,1.,.2);
  glutSolidSphere(l,20,20);
  glPopMatrix();
  glPushMatrix(); // #3
  glRotatef(90.-120.,0,1,0);
  glRotatef(a,0,0,1);
  glTranslatef(0,-l,0);
  glScalef(.2,1.,.2);
  glutSolidSphere(l,20,20);
  glPopMatrix();
  endinit();
}

avion::avion():stdmesh("avion",ICON_SCALE,.6*ICON_SCALE) {
  begininit();
  glTranslatef(0,0,1);
  glBegin(GL_TRIANGLES);
  glNormal3f(0,0,1);
  glVertex2f(-.6,-1);
  glVertex2f(-.1,-1);
  glVertex2f(  0, 1);
  glVertex2f( .6,-1);
  glVertex2f( .1,-1);
  glVertex2f(  0, 1);
  glNormal3f(.5,0,.4);
  glVertex3f(-.1,-1,0);
  glVertex3f(  0,-1,-.3);
  glVertex2f(  0, 1);
  glNormal3f(-.5,0,.4);
  glVertex3f( .1,-1,0);
  glVertex3f(  0,-1,-.3);
  glVertex2f(  0, 1);
  glEnd();
  endinit();
}

